@@ -14,8 +14,15 @@
 #include <R_ext/Rdynload.h>
 
 #define STRICT_R_HEADERS
+
+#ifndef R_NO_REMAP
 #define R_NO_REMAP
+#endif
+
+#ifndef R_USE_C99_IN_CXX
 #define R_USE_C99_IN_CXX
+#endif
+
 #include <R_ext/Error.h>
 
 #include <string>
@@ -831,7 +838,8 @@ SEXP GPB_CreateREModel_R(SEXP ndata,
 	SEXP likelihood,
 	SEXP likelihood_additional_param,
 	SEXP matrix_inversion_method,
-	SEXP seed) {
+	SEXP seed,
+	SEXP num_parallel_threads) {
 	SEXP ret;
 	REModelHandle handle = nullptr;
 	int32_t num_data = static_cast<int32_t>(Rf_asInteger(ndata));
@@ -881,6 +889,7 @@ SEXP GPB_CreateREModel_R(SEXP ndata,
 		Rf_asReal(likelihood_additional_param),
 		matrix_inversion_method_ptr,
 		Rf_asInteger(seed),
+		Rf_asInteger(num_parallel_threads),
 		&handle));
 	R_API_END();
 	ret = PROTECT(R_MakeExternalPtr(handle, R_NilValue, R_NilValue));
@@ -1318,7 +1327,7 @@ static const R_CallMethodDef CallEntries[] = {
   {"LGBM_BoosterSaveModel_R"          , (DL_FUNC)&LGBM_BoosterSaveModel_R          , 4},
   {"LGBM_BoosterSaveModelToString_R"  , (DL_FUNC)&LGBM_BoosterSaveModelToString_R  , 4},
   {"LGBM_BoosterDumpModel_R"          , (DL_FUNC)&LGBM_BoosterDumpModel_R          , 3},
-  {"GPB_CreateREModel_R"              , (DL_FUNC)&GPB_CreateREModel_R              , 27},
+  {"GPB_CreateREModel_R"              , (DL_FUNC)&GPB_CreateREModel_R              , 28},
   {"GPB_REModelFree_R"                , (DL_FUNC)&GPB_REModelFree_R                , 1},
   {"GPB_SetOptimConfig_R"             , (DL_FUNC)&GPB_SetOptimConfig_R             , 28},
   {"GPB_OptimCovPar_R"                , (DL_FUNC)&GPB_OptimCovPar_R                , 3},
