@@ -36,6 +36,8 @@ namespace GPBoost {
 		/*! \brief Virtual destructor */
 		virtual ~RECompBase() {};
 
+		virtual std::shared_ptr<RECompBase> clone() const = 0;
+
 		/*!
 		* \brief Create and adds the matrix Z_
 		*			Note: this is currently only used when changing the likelihood in the re_model
@@ -213,6 +215,21 @@ namespace GPBoost {
 		/*! \brief Constructor */
 		RECompGroup();
 
+		/*! \brief Copy constructor */
+		RECompGroup(const RECompGroup& other)
+			: RECompBase<T_mat>(other), // copy base class
+			num_group_(other.num_group_),
+			map_group_label_index_(std::make_shared<std::map<re_group_t, int>>(*other.map_group_label_index_)),
+			ZZt_(other.ZZt_),
+			has_ZZt_(other.has_ZZt_)
+		{
+			// No need to copy members of base class manually; base class copy constructor handles that.
+		}
+
+		std::shared_ptr<RECompBase<T_mat>> clone() const override {
+			return std::make_shared<RECompGroup<T_mat>>(*this);
+		}
+
 		/*!
 		* \brief Constructor without random coefficient data
 		* \param group_data Group data: factorial variable between 1 and the number of different groups
@@ -755,6 +772,30 @@ namespace GPBoost {
 		/*! \brief Constructor */
 		RECompGP();
 
+		/*! \brief Copy constructor */
+		RECompGP(const RECompGP& other)
+			: RECompBase<T_mat>(other),  // copy base members
+			coords_(other.coords_),
+			coords_ind_point_(other.coords_ind_point_),
+			dist_(other.dist_ ? std::make_shared<T_mat>(*other.dist_) : nullptr),
+			dist_saved_(other.dist_saved_),
+			coord_saved_(other.coord_saved_),
+			cov_function_(other.cov_function_ ? std::make_shared<CovFunction<T_mat>>(*other.cov_function_) : nullptr),
+			sigma_(other.sigma_),
+			sigma_defined_(other.sigma_defined_),
+			is_cross_covariance_IP_(other.is_cross_covariance_IP_),
+			num_random_effects_(other.num_random_effects_),
+			apply_tapering_(other.apply_tapering_),
+			apply_tapering_manually_(other.apply_tapering_manually_),
+			tapering_has_been_applied_(other.tapering_has_been_applied_),
+			has_compact_cov_fct_(other.has_compact_cov_fct_)
+		{
+		}
+
+		std::shared_ptr<RECompBase<T_mat>> clone() const override {
+			return std::make_shared<RECompGP<T_mat>>(*this);
+		}
+
 		/*!
 		* \brief Constructor for Gaussian process
 		* \param coords Coordinates (features) for Gaussian process
